@@ -62,6 +62,8 @@ type file struct {
 	objectInfo storage.ObjectInfo // Info in the context of the workspace.
 
 	ir                   *ir.File
+	irQuery              queries.IR
+	fileQuery            queries.File
 	referenceableSymbols map[ir.FullName]*symbol
 	referenceSymbols     []*symbol
 	symbols              []*symbol
@@ -97,6 +99,11 @@ func (f *file) Reset(ctx context.Context) {
 		f.workspace.Release()
 		f.workspace = nil
 	}
+	// Evict the query key if there is a query cached on the file. We cache the [queries.File]
+	// query since this allows the executor to evict all dependent queries, e.g. AST and IR.
+	if f.fileQuery.Path != "" {
+		f.lsp.queryExecutor.Evict(f.fileQuery.Key())
+	}
 	// Clear the file as nothing should use it after reset. This asserts that.
 	*f = file{}
 }
@@ -210,6 +217,14 @@ func (f *file) RefreshWorkspace(ctx context.Context) {
 //
 // This operation requires a workspace.
 func (f *file) RefreshIR(ctx context.Context) {
+	if f.objectInfo == nil {
+		// This can happen if the file that is opened is excluded from the workspace through
+		// includes/excludes in the build. In those cases, we ignore the file, the same way
+		// the rest of our tools would.
+		// In the future, we may want to rework this behavior in the LSP.
+		return
+	}
+
 	f.lsp.logger.Info(
 		"parsing IR for file",
 		slog.String("uri", string(f.uri)),
@@ -221,21 +236,46 @@ func (f *file) RefreshIR(ctx context.Context) {
 	// Opener creates a cached view of all files in the workspace.
 	pathToFiles := f.workspace.PathToFile()
 	files := make([]*file, 0, len(pathToFiles))
-	openerMap := make(map[string]*source.File, len(pathToFiles))
+	var evictQueryKeys []any
+
+	openerMap := f.lsp.opener.Get()
 	for path, file := range pathToFiles {
-		openerMap[path] = file.file
+		current := openerMap[path]
+		// If there is no entry for the current path or if the file content has changed, we
+		// update the opener and set a new query.
+		if current == nil || current.Text() != file.file.Text() {
+			openerMap[path] = file.file
+			// Add the query key for eviction if there is a query cached on the file. We cache
+			// the [queries.File] query since this allows the executor to evict all dependent
+			// queries, e.g. AST and IR.
+			if file.fileQuery.Path != "" {
+				evictQueryKeys = append(evictQueryKeys, file.fileQuery.Key())
+			}
+			file.fileQuery = queries.File{
+				Opener:      file.lsp.opener,
+				Path:        path,
+				ReportError: true, // [queries.AST] sets this to be true.
+			}
+			file.irQuery = queries.IR{
+				Opener:  file.lsp.opener,
+				Path:    file.objectInfo.Path(),
+				Session: file.lsp.irSession,
+			}
+		}
 		files = append(files, file)
 	}
-	opener := source.NewMap(openerMap)
-
-	session := new(ir.Session)
-	queries := xslices.Map(files, func(file *file) incremental.Query[*ir.File] {
-		return queries.IR{
-			Opener:  opener,
-			Path:    file.objectInfo.Path(),
-			Session: session,
+	// Remove paths that are no longer in the current workspace and evict stale query keys.
+	for path := range openerMap {
+		if _, ok := pathToFiles[path]; !ok {
+			delete(openerMap, path)
 		}
+	}
+	f.lsp.queryExecutor.Evict(evictQueryKeys...)
+
+	queries := xslices.Map(files, func(file *file) incremental.Query[*ir.File] {
+		return file.irQuery
 	})
+
 	results, diagnosticReport, err := incremental.Run(
 		ctx,
 		f.lsp.queryExecutor,
@@ -588,10 +628,12 @@ func (f *file) irToSymbols(irSymbol ir.Symbol) ([]*symbol, []*symbol) {
 		resolved = append(resolved, method)
 
 		input, _ := irSymbol.AsMethod().Input()
+		// Method input must be a single message type.
+		inputAST := irSymbol.AsMethod().AST().AsMethod().Signature.Inputs().At(0)
 		inputSym := &symbol{
 			ir:   irSymbol,
 			file: f,
-			span: irSymbol.AsMethod().AST().AsMethod().Signature.Inputs().Span(),
+			span: inputAST.RemovePrefixes().Span(), // We always strip prefixes in case of streaming.
 			kind: &reference{
 				def:      input.AST(), // Only messages can be method inputs and outputs
 				fullName: input.FullName(),
@@ -600,10 +642,12 @@ func (f *file) irToSymbols(irSymbol ir.Symbol) ([]*symbol, []*symbol) {
 		unresolved = append(unresolved, inputSym)
 
 		output, _ := irSymbol.AsMethod().Output()
+		// Method output must be a single message type.
+		outputAST := irSymbol.AsMethod().AST().AsMethod().Signature.Outputs().At(0)
 		outputSym := &symbol{
 			ir:   irSymbol,
 			file: f,
-			span: irSymbol.AsMethod().AST().AsMethod().Signature.Outputs().Span(),
+			span: outputAST.RemovePrefixes().Span(), // We always strip prefixes in case of streaming.
 			kind: &reference{
 				def:      output.AST(), // Only messages can be method inputs and outputs
 				fullName: output.FullName(),
@@ -684,12 +728,17 @@ func (f *file) messageToSymbolsHelper(msg ir.MessageValue, index int, parents []
 		//
 		// In the example, in the second definition, (option) and .message for field_b has a
 		// separate span from (option) and .message for field_a, but when we walk the mesasge
-		// tree, we get the span for (option) and .mesage for the first field. So we check the
+		// tree, we get the span for (option) and .message for the first field. So we check the
 		// symbols we've collected so far in parents and make sure we have captured a symbol for
 		// each path component.
 		for element := range seq.Values(field.Elements()) {
 			key := field.KeyASTs().At(element.ValueNodeIndex())
 			components := slices.Collect(key.AsPath().Components)
+			// If there are no path components for an element, then we skip it, since there are
+			// no symbols to track.
+			if len(components) == 0 {
+				continue
+			}
 			var span source.Span
 			// This covers the first case in the example above where the path is relative,
 			// e.g. field_a is a relative path within { } for (option).message.
@@ -854,36 +903,33 @@ func (f *file) RunChecks(ctx context.Context) {
 	f.cancelChecks = cancel
 
 	go func() {
-		image, diagnostics := buildImage(ctx, path, f.lsp.logger, opener)
-		if image == nil {
-			f.lsp.logger.DebugContext(ctx, "checks cancelled on image build", slog.String("uri", f.uri.Filename()))
-			return
-		}
-
-		f.lsp.logger.DebugContext(ctx, "checks running lint", slog.String("uri", f.uri.Filename()), slog.String("module", module.OpaqueID()))
 		var annotations []bufanalysis.FileAnnotation
-		if err := checkClient.Lint(
-			ctx,
-			workspace.GetLintConfigForOpaqueID(module.OpaqueID()),
-			image,
-			bufcheck.WithPluginConfigs(workspace.PluginConfigs()...),
-			bufcheck.WithPolicyConfigs(workspace.PolicyConfigs()...),
-		); err != nil {
-			var fileAnnotationSet bufanalysis.FileAnnotationSet
-			if !errors.As(err, &fileAnnotationSet) {
-				if errors.Is(err, context.Canceled) {
-					f.lsp.logger.DebugContext(ctx, "checks cancelled", slog.String("uri", f.uri.Filename()), xslog.ErrorAttr(err))
-				} else if errors.Is(err, context.DeadlineExceeded) {
-					f.lsp.logger.WarnContext(ctx, "checks deadline exceeded", slog.String("uri", f.uri.Filename()), xslog.ErrorAttr(err))
+		image, diagnostics := buildImage(ctx, path, f.lsp.logger, opener)
+		if image != nil {
+			f.lsp.logger.DebugContext(ctx, "checks running lint", slog.String("uri", f.uri.Filename()), slog.String("module", module.OpaqueID()))
+			if err := checkClient.Lint(
+				ctx,
+				workspace.GetLintConfigForOpaqueID(module.OpaqueID()),
+				image,
+				bufcheck.WithPluginConfigs(workspace.PluginConfigs()...),
+				bufcheck.WithPolicyConfigs(workspace.PolicyConfigs()...),
+			); err != nil {
+				var fileAnnotationSet bufanalysis.FileAnnotationSet
+				if !errors.As(err, &fileAnnotationSet) {
+					if errors.Is(err, context.Canceled) {
+						f.lsp.logger.DebugContext(ctx, "checks cancelled", slog.String("uri", f.uri.Filename()), xslog.ErrorAttr(err))
+					} else if errors.Is(err, context.DeadlineExceeded) {
+						f.lsp.logger.WarnContext(ctx, "checks deadline exceeded", slog.String("uri", f.uri.Filename()), xslog.ErrorAttr(err))
+					} else {
+						f.lsp.logger.WarnContext(ctx, "checks failed", slog.String("uri", f.uri.Filename()), xslog.ErrorAttr(err))
+					}
+					return
+				}
+				if len(fileAnnotationSet.FileAnnotations()) == 0 {
+					f.lsp.logger.DebugContext(ctx, "checks lint passed", slog.String("uri", f.uri.Filename()))
 				} else {
-					f.lsp.logger.WarnContext(ctx, "checks failed", slog.String("uri", f.uri.Filename()), xslog.ErrorAttr(err))
+					annotations = append(annotations, fileAnnotationSet.FileAnnotations()...)
 				}
-				return
-			}
-			if len(fileAnnotationSet.FileAnnotations()) == 0 {
-				f.lsp.logger.DebugContext(ctx, "checks lint passed", slog.String("uri", f.uri.Filename()))
-			} else {
-				annotations = append(annotations, fileAnnotationSet.FileAnnotations()...)
 			}
 		}
 
