@@ -307,6 +307,110 @@ func (s *symbol) GetSymbolInformation() protocol.SymbolInformation {
 	}
 }
 
+// Rename returns the [protocol.WorkspaceEdit] for renaming the symbol.
+func (s *symbol) Rename(newName string) (*protocol.WorkspaceEdit, error) {
+	var edits protocol.WorkspaceEdit
+	switch s.kind.(type) {
+	case *referenceable:
+		if err := checkRenameConflicts(s, newName); err != nil {
+			return nil, err
+		}
+		changes, err := renameChangesForReferenceableSymbol(s, newName)
+		if err != nil {
+			return nil, err
+		}
+		edits.Changes = changes
+	case *static:
+		if err := checkRenameConflicts(s, newName); err != nil {
+			return nil, err
+		}
+		edits.Changes = map[protocol.DocumentURI][]protocol.TextEdit{
+			s.file.uri: {{
+				Range:   reportSpanToProtocolRange(s.span),
+				NewText: newName,
+			}},
+		}
+	case *reference:
+		// For references, we attempt to rename the definition symbol, if resolved. This would
+		// include this reference symbol.
+		if s.def != nil {
+			if err := checkRenameConflicts(s.def, newName); err != nil {
+				return nil, err
+			}
+			changes, err := renameChangesForReferenceableSymbol(s.def, newName)
+			if err != nil {
+				return nil, err
+			}
+			edits.Changes = changes
+		}
+	}
+	// All other symbol types (options, imports, built-ins, and tags) cannot be renamed.
+	return &edits, nil
+}
+
+// renameChangesForReferenceableSymbol is a helper for getting all rename changes for the
+// given referenceable symbol.
+func renameChangesForReferenceableSymbol(s *symbol, newName string) (map[protocol.DocumentURI][]protocol.TextEdit, error) {
+	// At minimum, we would rename the symbol itself.
+	changes := map[protocol.DocumentURI][]protocol.TextEdit{
+		s.file.uri: {{
+			Range:   reportSpanToProtocolRange(s.span),
+			NewText: newName,
+		}},
+	}
+	if def, ok := s.def.kind.(*referenceable); ok {
+		for _, reference := range def.references {
+			changes[reference.file.uri] = append(changes[reference.file.uri], protocol.TextEdit{
+				Range:   reportSpanToProtocolRange(reference.span),
+				NewText: newName,
+			})
+		}
+	} else {
+		return nil, fmt.Errorf("attempting to rename a non-referenceble symbol as a referenceable symbol: %v", s)
+	}
+	return changes, nil
+}
+
+// checkRenameConflicts takes the symbol and desired new name and checks if this conflicts
+// with an existing symbol in the same scope and returns an error if a conflict is found.
+func checkRenameConflicts(target *symbol, newName string) error {
+	parent := target.ir.FullName().Parent()
+	if parent != "" {
+		var existing source.Span
+		newFullName := parent.Append(newName)
+		containsFunc := func(s *symbol) bool {
+			existing = s.span
+			return s.ir.FullName() == newFullName
+		}
+		// We check all files in the workspace for a conflict, since a package can span an arbitrary
+		// number of files.
+		// We first check the current symbol's file.
+		if slices.ContainsFunc(target.file.symbols, containsFunc) {
+			return fmt.Errorf(
+				"Renaming %q to %q would conflict with existing symbol at %s:%d:%d",
+				target.ir.FullName().Name(),
+				newName,
+				target.file.ir.Path(),
+				existing.StartLoc().Line,
+				existing.StartLoc().Column,
+			)
+		}
+		for _, file := range target.file.workspace.PathToFile() {
+			if slices.ContainsFunc(file.symbols, containsFunc) {
+				return fmt.Errorf(
+					"Renaming %q to %q would conflict with existing symbol at %s:%d:%d",
+					target.ir.FullName().Name(),
+					newName,
+					file.ir.Path(),
+					existing.StartLoc().Line,
+					existing.StartLoc().Column,
+				)
+			}
+		}
+	}
+	return nil
+}
+
 func protowireTypeForPredeclared(name predeclared.Name) protowire.Type {
 	switch name {
 	case predeclared.Bool, predeclared.Int32, predeclared.Int64, predeclared.UInt32,
@@ -353,22 +457,27 @@ func (s *symbol) getDocsFromComments() string {
 	_, start := def.Context().Stream().Around(def.Span().Start)
 	cursor := token.NewCursorAt(start)
 	t := cursor.PrevSkippable()
+	var addNewline bool
 	for !t.IsZero() {
 		switch t.Kind() {
 		case token.Comment:
-			comments = append(comments, commentToMarkdown(t.Text()))
+			text := t.Text()
+			if addNewline {
+				text += "\n"
+				addNewline = false
+			}
+			comments = append(comments, commentToMarkdown(text))
+		case token.Space:
+			// If the space token only contains spaces (e.g. code indentation), then we drop it.
+			// If the space token ends in a new-line, we append it to the comment above for formatting.
+			if strings.HasSuffix(t.Text(), "\n") {
+				addNewline = true
+			}
 		}
 		prev := cursor.PeekPrevSkippable()
 		if !prev.Kind().IsSkippable() {
 			break
 		}
-		if prev.Kind() == token.Space {
-			// Check if the whitespace contains a newline. If so, then we break. This is to prevent
-			// picking up comments that are not contiguous to the declaration.
-			if strings.Contains(prev.Text(), "\n") {
-				break
-			}
-		}
 		t = cursor.PrevSkippable()
 	}
 	comments = lineUpComments(comments)
